from typing import Callable, List, Tuple
from dateutil.parser import parse as date_parser
from datetime import datetime, timedelta, timezone
from operator import methodcaller
from functools import reduce
from isoduration import parse_duration

import pendulum

from . import monad

@monad.Try()
def try_parse_iso8601_time(iso_str: str):
    return pendulum.parse(iso_str)


@monad.Try()
def try_parse_iso8601_duration(iso_str: str):
    return parse_duration(iso_str)


def time_now(tz: str = "UTC", apply: List[Callable] = None) -> datetime:
    """
    Generates Time now as a DateTime object without args.

    Add a TZ by providing a TZ arg:
    > time_now(tz=tz_utc)

    Call an addition functions on the datetime using the apply arg.
    + To format in iso8601, provide the iso8601 function
    > time_now(tz=tz_utc, apply=[iso8601()]
    """
    dt = pendulum.now(tz=tz)
    if apply is None:
        return dt
    return apply_appliers(dt, apply)

def tz_convert(time, zone):
    """
    Converts a time from one TZ to another.
    The Zone is obtained from the tz() fn
    """
    return time.astimezone(zone)

def apply_appliers(obj, appliers: List[Callable]):
    return reduce(apply_applier, appliers, obj)

def apply_applier(obj, applier: Callable):
    return applier(obj)

def iso8601():
    """
    Generates a callable to convert a time into a ISO8601 Format
    """
    return methodcaller('isoformat')

def epoch():
    """
    Generates a callable to convert a time into a unix epoch timestamp
    """
    return methodcaller('timestamp')

def tz_utc():
    return 'UTC'

def tz(zone_name):
    return pendulum.timezone(zone_name)


def time_now_with_delta_seconds(delta):
    inc = timedelta(seconds=delta)
    return time_now(tz=timezone.utc) + inc

def time_with_delta(time=time_now(tz=timezone.utc), hours=0, minutes=0, seconds=0, direction='inc'):
    delta = timedelta(hours=hours, minutes=minutes, seconds=seconds)
    return time + delta if direction == 'inc' else time - delta


def iso_time_to_date_time(iso_time: str) -> datetime:
    return date_parser(iso_time)

def hours_from_start_from_year(time: datetime) -> int:
    return ((int(time.strftime("%j")) - 1) * 24) + time.hour

def hours_from_start_from_year_to_days_hours(start_time: datetime, hour: int) -> datetime:
    return time_with_delta(start_time, hours=hour)

def now_year_day() -> Tuple[int, int]:
    t = time_now(tz=timezone.utc)
    return (t.year, int(t.strftime("%j")))



class Singleton(object):
    _instance = None
    def __new__(self):
        if self._instance is None:
            self._instance = super(Singleton, self).__new__(self)
        return self._instance



from typing import List, Self
import traceback
from pymonad.operators.either import Either
from typing import TypeVar, Callable, Union, Any, Generic

T = TypeVar('T')


class EitherMonad(Generic[T]):
    pass


class MEither(Either):

    def __or__(self, fns):
        """
        Acts as a Monadic OR, determining the function to execute based on the Either.
        Provide 2 terminating functions (failure_fn, success_fn)
        """
        return self.either(fns[0], fns[1])

    # Lifts the Either; returning the wrapped value regardless of Left or Right
    def lift(self):
        return self.value if self.is_right() else self.monoid[0]

    def error(self):
        return self.monoid[0]


def Left(value: Either[T, T]) -> Either[T, T]:  # pylint: disable=invalid-name
    """ Creates a value of the first possible type in the Either monad. """
    return MEither(None, (value, False))


def Right(value: Either[T, T]) -> Either[T, T]:  # pylint: disable=invalid-name
    """ Creates a value of the second possible type in the Either monad. """
    return MEither(value, (None, True))


def Lift(value: Either[T, T]) -> T:  # pylint: disable=invalid-name
    return value.value if value.is_right() else value.error()


def maybe_value_ok(value: Either) -> bool:
    return value.is_right()


def maybe_value_fail(value: Either) -> bool:
    return value.is_left()


def monadic_try(name: str = None,
                status: int = None,
                exception_test_fn: Callable[[Either], Either] = None,
                error_cls: Any = None,
                error_result_fn: Callable = None):
    """
    Monadic Try Decorator.  Decorate any function which might return an exception.  When the function does not return an exception,
    the decorator wraps the result in a Right(), otherwise, it wraps the exception in a Left()

    Args:
        name: The name to be provided to the error object
        status: The status to be provided to the error object
        exception_test_fn: A function which takes the result.
                           When the result is Right, but the data may result in a Left.
                           The fn must return a value wrapped in an Either.
                           This fn can also be obtained from the expectation_fn to the main fn in the first instance
        error_result_fn:   A function whose result will be returned in the exception flow.  It takes a built exception (either str or error_cls).
                           If it takes an injected arg (error_result_fn_arg from main fn), it should be partially applied.

    The @monadic_try(name="step") is really syntax sugar for:
        $ monadic_try(name="x")(fn)(args)

        Hence the 3 nested functions
    """

    def inner(fn):
        def try_it(*args, **kwargs):
            try:
                result = Right(fn(*args, **kwargs))
                test_fn = kwargs.get('exception_test_fn', exception_test_fn)
                return test_fn(result) if test_fn else result
            except Exception as e:
                ex_cls = kwargs.get('error_cls', None) or error_cls

                error_result = Left(ex_cls(message=str(e),
                                           name=(kwargs.get('name', None) or name or fn.__name__),
                                           code=status,
                                           klass=str(e.__class__),
                                           exception_klass=e,
                                           traceback=traceback.format_exc())) if ex_cls else Left(str(e))

                return_fn = kwargs.get('error_result_fn', error_result_fn)
                if not return_fn:
                    return error_result

                injected_arg = kwargs.get('error_result_fn_arg', None)
                return return_fn(injected_arg, error_result)

        return try_it

    return inner


def aio_monadic_try(name: str = None,
                    status: int = None,
                    exception_test_fn: Callable[[Either], Either] = None,
                    error_cls: Any = None,
                    error_result_fn: Callable = None):
    def inner(f):
        async def try_it(*args, **kwargs):
            try:
                result = Right(await f(*args, **kwargs))
                test_fn = kwargs.get('exception_test_fn', exception_test_fn)
                return test_fn(result) if test_fn else result
            except Exception as e:
                ex_cls = kwargs.get('error_cls', None) or error_cls

                error_result = Left(ex_cls(message=str(e),
                                           name=(kwargs.get('name', None) or name or f.__name__),
                                           code=status, klass=str(e.__class__),
                                           request_kwargs=kwargs,
                                           traceback=traceback.format_exc())) if ex_cls else Left(str(e))

                return_fn = kwargs.get('error_result_fn', error_result_fn)
                if not return_fn:
                    return error_result

                injected_arg = kwargs.get('error_result_fn_arg', None)
                return return_fn(injected_arg, error_result)

        return try_it

    return inner


Try = monadic_try

AIOTry = aio_monadic_try


def any_error(try_result: Either) -> Union[str, None]:
    return try_result.either(lambda res: str(res.error()), lambda res: None)


def either_to_status(monad: EitherMonad):
    if not isinstance(monad, MEither):
        return "ok"
    return "ok" if monad.is_right() else "fail"



from typing import Dict, Tuple, List, Callable
from collections.abc import Iterable
from pymonad.maybe import *
from pymonad.tools import curry
# from pymonad.list import *
from pymonad.reader import Pipe
from functools import reduce, partial
from operator import iconcat
import re

import collections.abc

from . import monad, chronos

"""
Common functions
"""

def identity(arg):
    return arg

def deep_get(dict: Dict, path: List[str], default: Any=None) -> Any:
    """
    Takes a Dict, a path into the Dict as a list, and a default if nothing found, and returns the item at the path.
    Note, only works for a nested Dict, no Lists are supported.
    Example:
        $ deep_get({'a': {'b': 1}}, ['a', 'b'], 0)
    """
    if not isinstance(path, Iterable):
        return None
    fst, rst = first(path), rest(path)
    if fst not in dict: return default
    if (fst in dict) and len(rst) == 0: return dict[fst]
    return deep_get(dict[fst], rst, default)

def fst_rst(iterable: list) -> Tuple:
    return (first(iterable), rest(iterable))

def rest(iterable):
    return iterable[1:]

def partial_first(iterable):
    return first(iterable)

def first(iterable, default=None, key=None):
    """
    Return first element of `iterable` that evaluates true, else return None
    (or an optional default value).
    >>> first([0, False, None, [], (), 42])
    42
    >>> first([0, False, None, [], ()]) is None
    True
    >>> first([0, False, None, [], ()], default='ohai')
    'ohai'
    >>> import re
    >>> m = first(re.match(regex, 'abc') for regex in ['b.*', 'a(.*)'])
    >>> m.group(1)
    'bc'
    The optional `key` argument specifies a one-argument predicate function
    like that used for `filter()`.  The `key` argument, if supplied, must be
    in keyword form.  For example:
    >>> first([1, 1, 3, 4, 5], key=lambda x: x % 2 == 0)
    4
    """
    if key is None:
        for el in iterable:
            if el:
                return el
    else:
        for el in iterable:
            if key(el):
                return el
    return default

def find_by_type(type, iterable):
    return partial_filter(type_predicate(type), iterable).then(partial_first)

def find_by_predicate(predicate_fn: Callable, iterable: List):
    return partial_filter(predicate_fn, iterable).then(partial_first)

def type_predicate(type):
    return lambda x: x['_type'] == type

def partial_filter(fn: Callable, iterable: List):
    return Just(list(filter(fn, iterable)))

def find_by_filter(fn, xs):
    return next(filter(fn, xs), None)

def find(fn, xs):
    """
    fn.find(fn.equality(fn.at('a')), '1', [{'a': '1'}])
    """
    return next(select(fn, xs), None)

def select(fn: Callable, xs: List) -> filter:
    """
    returns all values from the list that are true when applying the fn to the item
    """
    return filter(fn, xs)

# + field_fn; the property to extract from the record.  Either a String or a Function which takes the record
# + test_value; the value which has == applied to determine equality
# + i; the record under test
# e.g. equality('a', "equal")({'a': "equal"})
# e.g. equality.(test_fn).("equal")({'a': "equal"})) ; where test_fn is -> x { x[:a] }
@curry(3)
def equality(field_or_fn, test_value, i):
    if callable(field_or_fn):
        return field_or_fn(i) == test_value
    else:
        return i[field_or_fn] == test_value


@curry(2)
def at(x, i):
    if x is None:
        return None

    elif x not in i:
        return None
    else:
        return i[x]

def match(pattern, test_string):
    return re.match(pattern, test_string)

# Curryed fn that removes elements from a collection where f.(e) is true
@curry(2)
def remove(fn: Callable, xs: List) -> List:
    return list(filter(partial(negated_fn,fn), xs))

def negated_fn(fn: Callable, x):
    return not fn(x)

def remove_none(xs):
    return list(filter(identity, xs))

def not_empty(xs: List[Any]) -> bool:
    return len(xs) > 0

def only_one(xs: List[Any]) -> bool:
    return len(xs) == 1

def bool_fn_with_predicate(xs, bool_fn, predicate):
    """
    Applies a bool fn over a list which is converted to bool through the predicate.
    The bool_fn can be anything which works with a list[bool]; like any, all
    > bool_fn_with_predicate([monad.Right(1), monad.Left(2)], any, monad.maybe_value.ok)
    """
    return bool_fn(map(predicate, xs))


def compose_iter(fn_list: List, initial_val):
    return reduce(lambda pipe, fn: pipe.then(fn), fn_list, Pipe(initial_val)).flush()


def either_compose(fn_list: List, initial_val):
    return reduce(lambda m, fn: m.bind(fn), fn_list, initial_val)

def flatten(xs: List):
    return reduce(iconcat, xs, [])



from typing import Dict, Tuple
from pymonad.tools import curry

@curry(2)
def at(attr, rec):
    return getattr(rec, attr, None)


@curry(2)
def at_prop(prop, rec):
    property = getattr(rec, prop, None)
    return property() if property and callable(property) else None

